# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from types import ModuleType

from distutils.version import StrictVersion

from neutron.plugins.ml2.drivers import type_tunnel
from neutron import version


# Some constants and verifier functions have been deprecated but are still
# used by earlier releases of neutron. In order to maintain
# backwards-compatibility with stable/mitaka this will act as a translator
# that passes constants and functions according to version number.

NEUTRON_VERSION = StrictVersion(str(version.version_info))
NEUTRON_NEWTON_VERSION = StrictVersion('9.0.0')
NEUTRON_OCATA_VERSION = StrictVersion('10.0.0')
NEUTRON_PIKE_VERSION = StrictVersion('11.0.0')
NEUTRON_QUEENS_VERSION = StrictVersion('12.0.0')

n_c = __import__('neutron.common.constants', fromlist=['common.constants'])
n_p_c = __import__('neutron.plugins.common.constants',
                   fromlist=['plugins.common.constants'])
n_p_c_attr_names = dir(n_p_c)
constants = __import__('neutron_lib.constants', fromlist=['constants'])

if NEUTRON_VERSION >= NEUTRON_NEWTON_VERSION:
    from neutron.conf import common as base_config
    from neutron_lib.api import validators
    from neutron_lib.db import model_base
    is_attr_set = validators.is_attr_set
    validators = validators.validators
    n_c_attr_names = getattr(n_c, "_mg__my_globals", None)
    HasProject = model_base.HasProject
else:
    from neutron.api.v2 import attributes
    from neutron.common import config as base_config
    n_c_attr_names = n_c.my_globals
    is_attr_set = attributes.is_attr_set
    validators = attributes.validators
    setattr(constants, 'ATTR_NOT_SPECIFIED', getattr(attributes,
                                                     'ATTR_NOT_SPECIFIED'))


if NEUTRON_VERSION >= NEUTRON_OCATA_VERSION:
    from neutron.agent.common import utils as agent_utils
    from neutron.db import api as db_api
    from neutron.db.models import agent as agent_model
    from neutron.db.models import l3 as l3_models
    from neutron.db.models import l3agent as rb_model
    from neutron.db import segments_db
    from neutron.objects import trunk as trunk_objects
    from neutron.services.trunk import constants as trunk_consts
    from neutron.services.trunk.drivers import base as trunk_base
    from neutron_lib.api.definitions import portbindings
    from neutron_lib.api.definitions import provider_net as providernet
    from neutron_lib.api import extensions
    from neutron_lib.plugins import directory
    from neutron_lib.services import base as service_base
    from neutron_lib.utils import helpers as common_utils

    try:
        from neutron import context
    except ImportError:
        from neutron_lib import context

    get_plugin = directory.get_plugin
    n_c_attr_names = dir(n_c)
    VXLAN_TUNNEL_TYPE = type_tunnel.ML2TunnelTypeDriver
    Agent = agent_model.Agent
    RouterPort = l3_models.RouterPort
    Router = l3_models.Router

    def get_context():
        return context.Context()

    def get_db_ref(context):
        return context

    def get_tunnel_session(context):
        return context.session

    def get_novaclient_images(nclient):
        return nclient.glance

    def get_reader_session():
        return db_api.get_reader_session()

    def get_writer_session():
        return db_api.get_writer_session()

    is_agent_down = agent_utils.is_agent_down

else:
    from networking_cisco.services.trunk import (  # noqa
        trunkstubs as trunk_objects)  # noqa
    from networking_cisco.services.trunk import (  # noqa
        trunkstubs as trunk_consts)  # noqa
    from networking_cisco.services.trunk import (  # noqa
        trunkstubs as trunk_base)  # noqa

    from neutron.api import extensions  # noqa
    from neutron.api.v2 import attributes as attr
    from neutron.common import utils as common_utils  # noqa
    from neutron import context
    from neutron.db import agents_db
    from neutron.db import api as db_api
    from neutron.db import l3_agentschedulers_db as rb_model  # noqa
    from neutron.db import l3_db
    from neutron.db import model_base  # noqa
    from neutron.extensions import portbindings  # noqa
    from neutron.extensions import providernet  # noqa
    from neutron import manager
    from neutron.plugins.common import constants as svc_constants
    from neutron.plugins.ml2 import db as segments_db  # noqa
    from neutron.services import service_base  # noqa
    import sqlalchemy as sa
    from sqlalchemy.ext import declarative
    from sqlalchemy import orm

    class HasTenant(object):

        project_id = sa.Column(sa.String(attr.TENANT_ID_MAX_LEN), index=True)

        def get_tenant_id(self):
            return self.project_id

        def set_tenant_id(self, value):
            self.project_id = value

        @declarative.declared_attr
        def tenant_id(cls):
            return orm.synonym(
                'project_id',
                descriptor=property(cls.get_tenant_id, cls.set_tenant_id))

    def get_plugin(service=None):
        if service is None:
            return manager.NeutronManager.get_plugin()
        else:
            return manager.NeutronManager.get_service_plugins().get(service)

    HasProject = HasTenant
    setattr(constants, 'L3', getattr(svc_constants, 'L3_ROUTER_NAT'))
    VXLAN_TUNNEL_TYPE = type_tunnel.TunnelTypeDriver
    Agent = agents_db.Agent
    RouterPort = l3_db.RouterPort
    Router = l3_db.Router

    def get_context():
        return context.Context(user_id=None, tenant_id=None)

    def get_db_ref(context):
        return db_api.get_session()

    def get_tunnel_session(context):
        return context

    def get_novaclient_images(nclient):
        return nclient.images

    def get_reader_session():
        return db_api.get_session()

    def get_writer_session():
        return db_api.get_session()

    is_agent_down = agents_db.AgentDbMixin.is_agent_down

if NEUTRON_VERSION >= NEUTRON_PIKE_VERSION:
    from neutron.conf.agent import common as config
    from neutron.db._resource_extend import extends
    from neutron.db._resource_extend import get_funcs  # noqa
    from neutron.db._resource_extend import has_resource_extenders  # noqa
    from neutron_lib.api.definitions import dns
    from neutron_lib.api.definitions import provider_net

    def auto_schedule_routers(self, hosts, r_ids):
        self.l3_plugin.auto_schedule_routers(self.adminContext, hosts)
else:
    from neutron.agent.common import config  # noqa
    from neutron.extensions import dns  # noqa
    from neutron.extensions import providernet as provider_net  # noqa

    def auto_schedule_routers(self, hosts, r_ids):
        self.l3_plugin.auto_schedule_routers(self.adminContext, hosts, r_ids)

    def has_resource_extenders(klass):
        return klass

    def extends(resources):
        def decorator(method):
            return method

        return decorator

if NEUTRON_VERSION >= NEUTRON_QUEENS_VERSION:
    # Newer than queens
    from neutron.conf.agent import common as neutron_agent_conf
    from neutron.conf.plugins.ml2 import config as ml2_config
    from neutron_lib.api.definitions import dns as dns_const
    from neutron_lib.api.definitions import external_net as exnet_const
    from neutron_lib.api.definitions import extraroute as extraroute_const
    from neutron_lib.api.definitions import l3 as l3_const
    from neutron_lib.api import faults as cb_faults
    from neutron_lib.callbacks import events as cb_events
    from neutron_lib.callbacks import registry as cb_registry
    from neutron_lib.callbacks import resources as cb_resources
    from neutron_lib import constants as cb_constants
    from neutron_lib.exceptions import agent as agent_exceptions
    from neutron_lib.exceptions import l3 as l3_exceptions
    from neutron_lib.plugins.ml2 import api as ml2_api
    from neutron_lib.utils import runtime as runtime_utils
    extraroute_const.EXTENDED_ATTRIBUTES_2_0 = (
            extraroute_const.RESOURCE_ATTRIBUTE_MAP)
    dns_const.EXTENDED_ATTRIBUTES_2_0 = dns_const.RESOURCE_ATTRIBUTE_MAP

    # Return the database object rather than oslo versioned object
    def get_agent_db_obj(agent):
        return agent.db_obj
else:
    # Pre-queens
    from neutron.agent.linux import external_process  # noqa
    from neutron.agent.linux import interface as neutron_agent_conf  # noqa
    from neutron.api.v2 import base as cb_faults  # noqa
    from neutron.callbacks import events as cb_events  # noqa
    from neutron.callbacks import registry as cb_registry  # noqa
    from neutron.callbacks import resources as cb_resources  # noqa
    from neutron.common import utils as runtime_utils  # noqa
    from neutron.extensions import agent as agent_exceptions  # noqa
    from neutron.extensions import dns as dns_const  # noqa
    from neutron.extensions import external_net as exnet_const  # noqa
    from neutron.extensions import extraroute as extraroute_const  # noqa
    from neutron.extensions import l3 as l3_const  # noqa
    from neutron.plugins.common import constants as cb_constants  # noqa
    from neutron.plugins.ml2 import config as ml2_config  # noqa
    from neutron.plugins.ml2 import driver_api as ml2_api  # noqa
    l3_exceptions = l3_const
    neutron_agent_conf.INTERFACE_OPTS = neutron_agent_conf.OPTS
    neutron_agent_conf.EXTERNAL_PROCESS_OPTS = external_process.OPTS

    def get_agent_db_obj(agent):
        return agent

core_opts = base_config.core_opts

# Bring in the union of all constants in neutron.common.constants
# and neutron_lib.constants. Handle any duplicates by using the
# values in neutron_lib.
#
# In the plugin code, replace the following imports:
#     from neutron.common import constants
#     from neutron_lib import constants
# with (something like this):
#     from networking_cisco import backward_compatibility as bc
# Then constants are referenced as shown in this example:
#     port['devide_owner'] = bc.constants.DEVICE_OWNER_ROUTER_INTF

ignore = frozenset(['__builtins__', '__doc__', '__file__', '__name__',
                    '__package__', '__path__', '__version__'])
for attr_name in n_c_attr_names:
    attr = getattr(n_c, attr_name)
    if attr_name in ignore or isinstance(attr, ModuleType):
        continue
    else:
        setattr(constants, attr_name, attr)
for attr_name in n_p_c_attr_names:
    attr = getattr(n_p_c, attr_name)
    if attr_name in ignore or isinstance(attr, ModuleType):
        continue
    else:
        setattr(constants, attr_name, attr)
del n_c, ignore, attr_name, attr



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import socket

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

import networking_cisco.plugins.cisco.common.cisco_constants as cc
from neutron.agent.linux import utils as linux_utils

from networking_cisco._i18n import _

import pprint


LOG = logging.getLogger(__name__)


STATUS_OPTS = [
    cfg.IntOpt('device_connection_timeout', default=30,
               help=_("Time in seconds for connecting to a hosting device")),
    cfg.IntOpt('hosting_device_dead_timeout', default=300,
               help=_("The time in seconds until a backlogged hosting device "
                      "is presumed dead. This value should be set up high "
                      "enough to recover from a period of connectivity loss "
                      "or high load when the device may not be responding.")),
]

cfg.CONF.register_opts(STATUS_OPTS, "cfg_agent")


def _is_pingable(ip):
    """Checks whether an IP address is reachable by pinging.

    Use linux utils to execute the ping (ICMP ECHO) command.
    Sends 5 packets with an interval of 0.2 seconds and timeout of 1
    seconds. Runtime error implies unreachability else IP is pingable.
    :param ip: IP to check
    :return: bool - True or False depending on pingability.
    """
    ping_cmd = ['ping',
                '-c', '5',
                '-W', '1',
                '-i', '0.2',
                ip]
    try:
        linux_utils.execute(ping_cmd, check_exit_code=True)
        return True
    except RuntimeError:
        LOG.warning("Cannot ping ip address: %s", ip)
        return False


def _can_connect(ip, port):
    """Checks if a TCP port at IP address is possible to connect to"""
    cs = socket.socket()
    try:
        cs.connect((ip, port))
        cs.close()
        return True
    except socket.error:
        return False


class DeviceStatus(object):
    """Device status and backlog processing."""

    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(DeviceStatus, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self.hosting_devices_backlog = {}
        self.enable_heartbeat = False

    def backlog_hosting_device(self, hosting_device):
        hd_id = hosting_device['id']
        hd_mgmt_ip = hosting_device['management_ip_address']
        # Modifying the 'created_at' to a date time object if it is not
        if not isinstance(hosting_device['created_at'], datetime.datetime):
            hosting_device['created_at'] = datetime.datetime.strptime(
                hosting_device['created_at'], '%Y-%m-%d %H:%M:%S')
        if hd_id not in self.hosting_devices_backlog:
            if 'hd_state' not in hosting_device:
                hosting_device['hd_state'] = hosting_device['status']
            hosting_device['backlog_insertion_ts'] = max(
                timeutils.utcnow(),
                hosting_device['created_at'] +
                datetime.timedelta(seconds=hosting_device['booting_time']))
            self.hosting_devices_backlog[hd_id] = {'hd': hosting_device}
            LOG.debug("Hosting device: %(hd_id)s @ %(ip)s is now added "
                      "to backlog", {'hd_id': hd_id, 'ip': hd_mgmt_ip})

    def backlog_hosting_devices(self, hosting_devices):
        for hosting_device in hosting_devices:
            self.backlog_hosting_device(hosting_device)

    def get_backlogged_hosting_devices(self):
        return self.hosting_devices_backlog.keys()

    def get_backlogged_hosting_devices_info(self):
        resp = self.get_monitored_hosting_devices_info(hd_state_filter='Dead')
        return resp

    def get_dead_hosting_devices_info(self):
        """
        Get a list of hosting devices that have been marked dead
        :return: List of dead hosting device ids
        """
        res = []
        for hd_id in self.hosting_devices_backlog:
            hd = self.hosting_devices_backlog[hd_id]['hd']
            if hd['hd_state'] == cc.HD_DEAD:
                res.append(hd['id'])
        return res

    def get_monitored_hosting_devices_info(self, hd_state_filter=None):
        """
        This function returns a list of all hosting devices monitored
        by this agent
        """
        wait_time = datetime.timedelta(
            seconds=cfg.CONF.cfg_agent.hosting_device_dead_timeout)
        resp = []
        for hd_id in self.hosting_devices_backlog:
            hd = self.hosting_devices_backlog[hd_id]['hd']

            display_hd = True

            if hd_state_filter is not None:
                if hd['hd_state'] == hd_state_filter:
                    display_hd = True
                else:
                    display_hd = False

            if display_hd:
                created_time = hd['created_at']
                boottime = datetime.timedelta(seconds=hd['booting_time'])
                backlogged_at = hd['backlog_insertion_ts']
                booted_at = created_time + boottime
                dead_at = backlogged_at + wait_time
                resp.append({'host id': hd['id'],
                             'hd_state': hd['hd_state'],
                             'created at': str(created_time),
                             'backlogged at': str(backlogged_at),
                             'estimate booted at': str(booted_at),
                             'considered dead at': str(dead_at)})
            else:
                continue
        return resp

    def is_hosting_device_reachable(self, hosting_device):
        """Check the hosting device which hosts this resource is reachable.

        If the resource is not reachable, it is added to the backlog.

        * heartbeat revision

        We want to enqueue all hosting-devices into the backlog for
        monitoring purposes

        adds key/value pairs to  hd (aka hosting_device dictionary)

        _is_pingable : if it returns true,
            hd['hd_state']='Active'
        _is_pingable : if it returns false,
            hd['hd_state']='Unknown'

        :param hosting_device : dict of the hosting device
        :returns: True if device is reachable, else None
        """
        ret_val = False

        hd = hosting_device
        hd_id = hosting_device['id']
        hd_mgmt_ip = hosting_device['management_ip_address']

        dead_hd_list = self.get_dead_hosting_devices_info()
        if hd_id in dead_hd_list:
            LOG.debug("Hosting device: %(hd_id)s@%(ip)s is already marked as"
                      " Dead. It is assigned as non-reachable",
                      {'hd_id': hd_id, 'ip': hd_mgmt_ip})
            return False

        # Modifying the 'created_at' to a date time object if it is not
        if not isinstance(hd['created_at'], datetime.datetime):
            hd['created_at'] = datetime.datetime.strptime(hd['created_at'],
                                                          '%Y-%m-%d %H:%M:%S')

        if _is_pingable(hd_mgmt_ip):
            LOG.debug("Hosting device: %(hd_id)s@%(ip)s is reachable.",
                      {'hd_id': hd_id, 'ip': hd_mgmt_ip})
            hd['hd_state'] = cc.HD_ACTIVE
            ret_val = True
        else:
            LOG.debug("Hosting device: %(hd_id)s@%(ip)s is NOT reachable.",
                      {'hd_id': hd_id, 'ip': hd_mgmt_ip})
            hd['hd_state'] = cc.HD_NOT_RESPONDING
            ret_val = False

        if self.enable_heartbeat is True or ret_val is False:
            self.backlog_hosting_device(hd)

        return ret_val

    def check_backlogged_hosting_devices(self, driver_mgr):
        """Checks the status of backlogged hosting devices.

        Skips newly spun up instances during their booting time as specified
        in the boot time parameter.

        Each hosting-device tracked has a key, hd_state, that represents the
        last known state for the hosting device.  Valid values for hd_state
        are ['Active', 'Unknown', 'Dead']

        Each time check_backlogged_hosting_devices is invoked, a ping-test
        is performed to determine the current state.  If the current state
        differs, hd_state is updated.

        The hd_state transitions/actions are represented by the following
        table.

        +------------+---------------------+----------------+----------------+
        | current /  | Active              | Unknown        | Dead           |
        | last state |                     |                |                |
        +============+=====================+================+================+
        | Active     | Device is reachable.| Device was     | Dead device    |
        |            | No state change     | temporarily    | recovered.     |
        |            |                     | unreachable.   | Trigger resync |
        +------------+---------------------+----------------+----------------+
        | Unknown    | Device connectivity | Device         | Not a valid    |
        |            | test failed.  Set   | connectivity   | state          |
        |            | backlog timestamp   | test failed.   | transition.    |
        |            | and wait for dead   | Dead timeout   |                |
        |            | timeout to occur.   | has not        |                |
        |            |                     | occurred yet.  |                |
        +------------+---------------------+----------------+----------------+
        | Dead       | Not a valid state   | Dead timeout   | Device is      |
        |            | transition.         | for device has | still dead.    |
        |            |                     | elapsed.       | No state       |
        |            |                     | Notify plugin  | change.        |
        +------------+---------------------+----------------+----------------+

        :returns: A dict of the format:

        ::

          {"reachable": [<hd_id>,..],
          "dead": [<hd_id>,..],
          "revived": [<hd_id>,..]}

        * reachable - a list of hosting devices that are now reachable
        * dead      - a list of hosting devices deemed dead
        * revived   - a list of hosting devices (dead to active)
        """
        response_dict = {'reachable': [], 'revived': [], 'dead': []}
        LOG.debug("Current Backlogged hosting devices: \n%s\n",
                  self.hosting_devices_backlog.keys())
        for hd_id in self.hosting_devices_backlog.keys():
            hd = self.hosting_devices_backlog[hd_id]['hd']
            if not timeutils.is_older_than(hd['created_at'],
                                           hd['booting_time']):
                LOG.info("Hosting device: %(hd_id)s @ %(ip)s hasn't "
                         "passed minimum boot time. Skipping it. ",
                         {'hd_id': hd_id, 'ip': hd['management_ip_address']})
                continue
            LOG.info("Checking hosting device: %(hd_id)s @ %(ip)s for "
                     "reachability.", {'hd_id': hd_id,
                                       'ip': hd['management_ip_address']})
            hd_state = hd['hd_state']
            if _is_pingable(hd['management_ip_address']):
                if hd_state == cc.HD_NOT_RESPONDING:
                    LOG.debug("hosting devices revived & reachable, %s" %
                              (pprint.pformat(hd)))
                    hd['hd_state'] = cc.HD_ACTIVE
                    # hosting device state
                    response_dict['reachable'].append(hd_id)
                elif hd_state == cc.HD_DEAD:
                    # test if management port is available
                    if _can_connect(hd['management_ip_address'],
                                    hd['protocol_port']) is True:
                        LOG.debug("Dead hosting devices revived %s" %
                                  (pprint.pformat(hd)))
                        hd['hd_state'] = cc.HD_ACTIVE
                        response_dict['revived'].append(hd_id)
                    else:
                        LOG.debug("Cannot connect to management port %(p)d on "
                                  "hosting device with ip %(ip)s",
                                  {'p': hd['protocol_port'],
                                   'ip': hd['management_ip_address']})
                else:
                    LOG.debug("No-op."
                              "_is_pingable is True and current"
                              " hd['hd_state']=%s" % hd_state)

                LOG.info("Hosting device: %(hd_id)s @ %(ip)s is now "
                         "reachable. Adding it to response",
                         {'hd_id': hd_id, 'ip': hd['management_ip_address']})
            else:
                LOG.info("Hosting device: %(hd_id)s %(hd_state)s"
                         " @ %(ip)s not reachable ",
                         {'hd_id': hd_id,
                          'hd_state': hd['hd_state'],
                          'ip': hd['management_ip_address']})
                if hd_state == cc.HD_ACTIVE:
                    LOG.debug("hosting device lost connectivity, %s" %
                              (pprint.pformat(hd)))
                    hd['backlog_insertion_ts'] = timeutils.utcnow()
                    hd['hd_state'] = cc.HD_NOT_RESPONDING

                elif hd_state == cc.HD_NOT_RESPONDING:
                    if timeutils.is_older_than(
                            hd['backlog_insertion_ts'],
                            cfg.CONF.cfg_agent.hosting_device_dead_timeout):
                        # current hd_state is now dead, previous state: Unknown
                        hd['hd_state'] = cc.HD_DEAD
                        LOG.debug("Hosting device: %(hd_id)s @ %(ip)s hasn't "
                                  "been reachable for the "
                                  "last %(time)d seconds. "
                                  "Marking it dead.",
                                  {'hd_id': hd_id,
                                   'ip': hd['management_ip_address'],
                                   'time': cfg.CONF.cfg_agent.
                                   hosting_device_dead_timeout})
                        response_dict['dead'].append(hd_id)
        LOG.debug("Response: %s", response_dict)
        return response_dict



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db import models_v2

from networking_cisco import backwards_compatibility as bc


class HostingDeviceTemplate(bc.model_base.BASEV2, bc.model_base.HasId,
                            bc.HasProject):
    """Represents a template for devices used to host service.

       Such devices may be physical or virtual.
    """
    __tablename__ = 'cisco_hosting_device_templates'

    # name given to hosting devices created using this template
    name = sa.Column(sa.String(255))
    # template enabled if True
    enabled = sa.Column(sa.Boolean, nullable=False,
                        server_default=sa.sql.false())
    # 'host_category' can be 'VM', 'Hardware', 'NetworkNode'
    host_category = sa.Column(sa.String(255), nullable=False)
    # list of service types hosting devices based on this template support
    service_types = sa.Column(sa.String(255))
    # the image name or UUID in Glance
    image = sa.Column(sa.String(255))
    # the VM flavor name or UUID in Nova
    flavor = sa.Column(sa.String(255))
    # UUID of default credentials (if any) for devices created from this
    # template
    default_credentials_id = sa.Column(sa.String(36))
    #TODO(bobmel): Evolve to support multiple mechanisms and protocol ports
    # 'configuration_mechanism' indicates how configurations are made
    configuration_mechanism = sa.Column(sa.String(255))
    # 'protocol_port' is udp/tcp port of hosting device. May be empty.
    protocol_port = sa.Column(sa.Integer)
    # Typical time (in seconds) needed for hosting device (created
    # from this template) to boot into operational state.
    booting_time = sa.Column(sa.Integer, server_default='0')
    # abstract metric specifying capacity to host logical resources
    slot_capacity = sa.Column(sa.Integer, nullable=False, autoincrement=False)
    # desired number of slots to keep available at all times
    desired_slots_free = sa.Column(sa.Integer, nullable=False,
                                   server_default='0', autoincrement=False)
    # 'tenant_bound' is a (possibly empty) string of ':'-separated tenant UUIDs
    # representing the only tenants allowed to own/place resources on
    # hosting devices created using this template. If string is empty all
    # tenants are allowed.
    tenant_bound = sa.Column(sa.String(512))
    # module to be used as hosting device driver when creating
    # hosting devices using his template
    device_driver = sa.Column(sa.String(255), nullable=False)
    # module to be used as plugging driver for logical resources
    # hosted inside hosting devices created using this template
    plugging_driver = sa.Column(sa.String(255), nullable=False)


class HostingDevice(bc.model_base.BASEV2, bc.model_base.HasId,
                    bc.HasProject):
    """Represents an appliance hosting Neutron router(s).

       When the hosting device is a Nova VM 'id' is uuid of that VM.
    """
    __tablename__ = 'cisco_hosting_devices'

    # complementary id to enable identification of associated Neutron resources
    complementary_id = sa.Column(sa.String(36))
    # UUID of hosting device template used to create the hosting device
    template_id = sa.Column(sa.String(36),
                            sa.ForeignKey('cisco_hosting_device_templates.id'),
                            nullable=False)
    template = orm.relationship(HostingDeviceTemplate)
    # UUID of credentials for this hosting device
    credentials_id = sa.Column(sa.String(36))
    # name of hosting device (e.g., its name in DNS)
    name = sa.Column(sa.String(255))
    # arbitrary description string
    description = sa.Column(sa.String(255))
    # manufacturer id of the device, e.g., its serial number
    device_id = sa.Column(sa.String(255))
    # version 4 or 6 IP address of management interface
    management_ip_address = sa.Column(sa.String(255))
    admin_state_up = sa.Column(sa.Boolean, nullable=False)
    # 'management_port_id' is the Neutron Port used for management interface
    management_port_id = sa.Column(sa.String(36),
                                   sa.ForeignKey('ports.id',
                                                 ondelete="SET NULL"))
    management_port = orm.relationship(models_v2.Port)
    #TODO(bobmel): Evolve to support multiple mgmt_ports, protocol ports
    # 'protocol_port' is udp/tcp port of hosting device. May be empty.
    protocol_port = sa.Column(sa.Integer)
    cfg_agent_id = sa.Column(sa.String(36),
                             sa.ForeignKey('agents.id'),
                             nullable=True)
    cfg_agent = orm.relationship(bc.Agent)
    # Service VMs take time to boot so we store creation time
    # so we can give preference to older ones when scheduling
    created_at = sa.Column(sa.DateTime, nullable=False)
    status = sa.Column(sa.String(16))
    # 'tenant_bound' is empty or is UUID of the only tenant allowed to
    # own/place resources on this hosting device
    tenant_bound = sa.Column(sa.String(36))
    # If 'auto_delete' is True, a VM-based hosting device is subject to
    # deletion as part of hosting device pool management and in case of VM
    # failures. If 'auto_delete' is set to False, the hosting device must be
    # manually unregistered in the device manager and deleted in Nova.
    auto_delete = sa.Column(sa.Boolean, nullable=False,
                            server_default=sa.sql.false())


class SlotAllocation(bc.model_base.BASEV2):
    """Tracks allocation of slots in hosting devices."""
    __tablename__ = 'cisco_slot_allocations'

    template_id = sa.Column(sa.String(36),
                            sa.ForeignKey('cisco_hosting_device_templates.id'),
                            nullable=False)
    hosting_device_id = sa.Column(sa.String(36),
                                  sa.ForeignKey('cisco_hosting_devices.id'),
                                  nullable=False)
    # type of logical resource (e.g., router)
    logical_resource_type = sa.Column(sa.String(155), nullable=False)
    # service managing the logical resource (e.g., L3_ROUTER_NAT)
    logical_resource_service = sa.Column(sa.String(155), nullable=False)
    logical_resource_id = sa.Column(sa.String(36), primary_key=True,
                                    nullable=False)
    # UUID of tenant owning logical resource
    logical_resource_owner = sa.Column(sa.String(36), nullable=False)
    num_allocated = sa.Column(sa.Integer, autoincrement=False, nullable=False)
    tenant_bound = sa.Column(sa.String(36))


class HostedHostingPortBinding(bc.model_base.BASEV2):
    """Represents binding of logical resource's port to its hosting port."""
    __tablename__ = 'cisco_port_mappings'

    logical_resource_id = sa.Column(sa.String(36), primary_key=True)
    logical_port_id = sa.Column(sa.String(36),
                                sa.ForeignKey('ports.id',
                                              ondelete="CASCADE"),
                                primary_key=True)
    logical_port = orm.relationship(
        models_v2.Port,
        primaryjoin='Port.id==HostedHostingPortBinding.logical_port_id',
        backref=orm.backref('hosting_info', cascade='all', uselist=False))
    # type of router port: router_interface, ..._gateway, ..._floatingip
    port_type = sa.Column(sa.String(32))
    # type of network the router port belongs to
    network_type = sa.Column(sa.String(32))
    hosting_port_id = sa.Column(sa.String(36),
                                sa.ForeignKey('ports.id',
                                              ondelete='CASCADE'))
    hosting_port = orm.relationship(
        models_v2.Port,
        primaryjoin='Port.id==HostedHostingPortBinding.hosting_port_id')
    # VLAN tag for trunk ports
    segmentation_id = sa.Column(sa.Integer, autoincrement=False)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from oslo_log import log as logging

from networking_cisco._i18n import _

from six.moves import queue as Queue

LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class ServiceHelperBase(object):

    def __init__(self):
        self._observers = []

    def register(self, observer):
        LOG.debug("Attaching observer: %(ob)s to subject: %(sub)s",
                  {'ob': observer.__class__.__name__,
                   'sub': self.__class__.__name__})
        if observer not in self._observers:
            self._observers.append(observer)
        else:
            raise ValueError(_("Observer: %(ob)s is already registered to "
                             "subject: %(sub)s"),
                             {'ob': observer.__class__.__name__,
                              'sub': self.__class__.__name__})

    def unregister(self, observer):
        LOG.debug("Dettaching observer: %(ob)s from subject: %(sub)s",
                  {'ob': observer.__class__.__name__,
                   'sub': self.__class__.__name__})
        if observer in self._observers:
            self._observers.remove(observer)
        else:
            raise ValueError(_("Observer: %(ob)s is not attached to "
                               "subject: %(sub)s"),
                             {'ob': observer.__class__.__name__,
                              'sub': self.__class__.__name__})

    def notify(self, resource, **kwargs):
        """Calls all observers attached to the given subject."""
        LOG.debug("Notifying all observers of this subject")
        for observer in self._observers:
            LOG.debug("Notifying observer: %s", observer.__class__.__name__)
            observer.update(resource, **kwargs)

    def update(self, resource, **kwargs):
        """For future support."""
        LOG.debug("Update received")

    @abc.abstractmethod
    def process_service(self, *args, **kwargs):
        raise NotImplementedError


class QueueMixin(object):
    def __init__(self):
        super(QueueMixin, self).__init__()
        self._queues = {}

    def enqueue(self, qname, data):
        if qname not in self._queues:
            self._queues[qname] = Queue.Queue()
        queue = self._queues[qname]
        queue.put(data)

    def dequeue(self, qname):
        if qname not in self._queues:
            raise ValueError(_("queue %s is not defined"), qname)
        try:
            return self._queues[qname].get(block=False)
        except Queue.Empty:
            return None

    def qsize(self, qname):
        """Return the approximate size of the queue."""
        if qname in self._queues:
            return self._queues[qname].qsize()
        else:
            raise ValueError(_("queue %s is not defined"), qname)



# Copyright (c) 2014-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import constants as p_const
from networking_cisco.backwards_compatibility import ml2_api as api

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import type_nexus_vxlan

from neutron.tests.unit import testlib_api

VNI_RANGES = [(100, 102), (200, 202)]
VNI_RANGE_LOW_INVALID = [str(type_nexus_vxlan.MIN_NEXUS_VNI - 1) + ':' +
                         str(p_const.MAX_VXLAN_VNI)]
VNI_RANGE_HIGH_INVALID = [str(type_nexus_vxlan.MIN_NEXUS_VNI) + ':' +
                          str(p_const.MAX_VXLAN_VNI + 1)]
MCAST_GROUP_RANGES = ['224.0.0.1:224.0.0.2', '224.0.1.1:224.0.1.2']


class NexusVxlanTypeTest(testlib_api.SqlTestCase):

    def setUp(self):
        super(NexusVxlanTypeTest, self).setUp()
        self.driver = type_nexus_vxlan.NexusVxlanTypeDriver()
        self.driver.conf_mcast_ranges = MCAST_GROUP_RANGES
        self.driver.tunnel_ranges = VNI_RANGES
        self.driver.sync_allocations()
        self.context = bc.get_db_ref(bc.get_context())

    def vni_in_range(self, vni):
        # SegmentTypeDriver.allocate_partially_specified_segment allocates
        # a random VNI from the range
        return any(lower <= vni <= upper for (lower, upper) in VNI_RANGES)

    def test_allocate_tenant_segment(self):
        segment = self.driver.allocate_tenant_segment(self.context)
        self.assertEqual(segment[api.NETWORK_TYPE], const.TYPE_NEXUS_VXLAN)
        self.assertEqual(segment[api.PHYSICAL_NETWORK], '224.0.0.1')
        self.assertTrue(self.vni_in_range(segment[api.SEGMENTATION_ID]))

    def test_allocate_shared_mcast_group(self):
        segments = []
        for i in range(0, 6):
            segments.append(self.driver.allocate_tenant_segment(
                self.context))
        self.assertEqual(segments[0][api.NETWORK_TYPE],
                         const.TYPE_NEXUS_VXLAN)
        self.assertEqual(segments[0][api.PHYSICAL_NETWORK], '224.0.0.1')
        self.assertTrue(self.vni_in_range(segments[0][api.SEGMENTATION_ID]))
        self.assertEqual(segments[-1][api.NETWORK_TYPE],
                         const.TYPE_NEXUS_VXLAN)
        self.assertEqual(segments[-1][api.PHYSICAL_NETWORK], '224.0.0.1')
        self.assertTrue(self.vni_in_range(segments[-1][api.SEGMENTATION_ID]))
        self.assertNotEqual(segments[0], segments[-1])

    def test_reserve_provider_segment_full_specs(self):
        segment = {api.NETWORK_TYPE: const.TYPE_NEXUS_VXLAN,
                   api.PHYSICAL_NETWORK: '224.0.0.1',
                   api.SEGMENTATION_ID: '5000'}
        result = self.driver.reserve_provider_segment(self.context, segment)
        alloc = self.driver.get_allocation(
            self.context,
            result[api.SEGMENTATION_ID])
        mcast_group = self.driver._get_mcast_group_for_vni(
            self.context,
            alloc.vxlan_vni)
        self.assertTrue(alloc.allocated)
        self.assertEqual(alloc.vxlan_vni, 5000)
        self.assertEqual(mcast_group, '224.0.0.1')

    def test_reserve_provider_segment_partial_specs(self):
        segment = {api.NETWORK_TYPE: const.TYPE_NEXUS_VXLAN,
                   api.PHYSICAL_NETWORK: '224.0.0.1'}

        result = self.driver.reserve_provider_segment(self.context, segment)
        alloc = self.driver.get_allocation(
            self.context,
            result[api.SEGMENTATION_ID])
        mcast_group = self.driver._get_mcast_group_for_vni(
            self.context,
            alloc.vxlan_vni)
        self.assertTrue(alloc.allocated)
        self.assertTrue(self.vni_in_range(alloc.vxlan_vni))
        self.assertEqual(mcast_group, '224.0.0.1')

    def test_invalid_vni_ranges(self):
        for invalid_vni_range in [VNI_RANGE_LOW_INVALID,
                                  VNI_RANGE_HIGH_INVALID]:
            type_nexus_vxlan.cfg.CONF.set_override('vni_ranges',
                                                   invalid_vni_range,
                                                   'ml2_type_nexus_vxlan')
            self.assertRaises(SystemExit, self.driver._verify_vni_ranges)



# Copyright (c) 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.nexus import trunk

from neutron.tests.unit.db import test_db_base_plugin_v2


PORT_ID = 'fake_port_id'
TRUNK_ID = 'fake_trunk_id'
DNS_NAME = 'test_dns_name'
VM_NAME = 'test_vm_name'
SEGMENTATION_VLAN = 'vlan'
SEGMENTATION_ID1 = 101
SEGMENTATION_ID2 = 102

SUBPORTS = [
    {'segmentation_type': SEGMENTATION_VLAN, 'port_id': PORT_ID,
     'segmentation_id': SEGMENTATION_ID1},
    {'segmentation_type': SEGMENTATION_VLAN, 'port_id': PORT_ID,
     'segmentation_id': SEGMENTATION_ID2}]

TRUNK = {
    'status': bc.constants.PORT_STATUS_ACTIVE,
    'sub_ports': SUBPORTS,
    'name': 'trunk0',
    'admin_state_up': 'true',
    'tenant_id': 'fake_tenant_id',
    'project_id': 'fake_project_id',
    'port_id': PORT_ID,
    'id': TRUNK_ID,
    'description': 'fake trunk port'}

PROFILE_BAREMETAL = [{"switch_info": "test_value"}]

SUBPORT = {
    'status': bc.constants.PORT_STATUS_ACTIVE,
    'port_id': PORT_ID,
    'segmentation_id': SEGMENTATION_ID1}

PORT_BAREMETAL = {
    'status': bc.constants.PORT_STATUS_ACTIVE,
    'id': PORT_ID,
    bc.portbindings.VNIC_TYPE: bc.portbindings.VNIC_BAREMETAL,
    bc.dns.DNSNAME: DNS_NAME,
    bc.portbindings.PROFILE: {"local_link_information": PROFILE_BAREMETAL},
    'trunk_details': {'trunk_id': TRUNK_ID, 'sub_ports': SUBPORTS}}

PORT_VM = {
    'status': bc.constants.PORT_STATUS_ACTIVE,
    'id': PORT_ID,
    bc.portbindings.VNIC_TYPE: bc.portbindings.VNIC_NORMAL,
    bc.portbindings.HOST_ID: VM_NAME,
    bc.portbindings.PROFILE: {},
    'trunk_details': {'trunk_id': TRUNK_ID, 'sub_ports': SUBPORTS}}


class TestSubPort(object):
    port_id = PORT_ID
    trunk_id = TRUNK_ID
    segmentation_type = SEGMENTATION_VLAN
    segmentation_id = SEGMENTATION_ID1


class TestTrunk(object):
    admin_state_up = 'test_admin_state'
    id = TRUNK_ID
    tenant_id = 'test_tenant_id'
    name = 'test_trunk_name'
    port_id = PORT_ID
    status = bc.constants.PORT_STATUS_ACTIVE
    sub_ports = SUBPORTS
    update = mock.Mock()


@testtools.skipIf(bc.NEUTRON_VERSION < bc.NEUTRON_OCATA_VERSION,
                  "Test not applicable prior to stable/ocata.")
class TestNexusTrunkHandler(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    def setUp(self):
        super(TestNexusTrunkHandler, self).setUp()

        self.handler = trunk.NexusMDTrunkHandler()
        self.plugin = bc.get_plugin()
        self.plugin.get_port = mock.Mock()
        self.plugin.update_port = mock.Mock()
        self.mock_subport_get_object = mock.patch.object(
            bc.trunk_objects.SubPort, 'get_object',
            return_value=TestSubPort).start()
        self.mock_trunk_get_object = mock.patch.object(
            bc.trunk_objects.Trunk, 'get_object',
            return_value=TestTrunk).start()
        self.mock_trunk_get_object = mock.patch.object(
            bc.trunk_objects.Trunk, 'get_object').start()

    def _test_update_subports(self, port, host_id):
        self.handler.update_subports(port)

        self.assertEqual(2, self.plugin.update_port.call_count)
        self.plugin.update_port.assert_called_with(mock.ANY, PORT_ID,
            {'port':
             {bc.portbindings.HOST_ID: host_id,
              'device_owner': bc.trunk_consts.TRUNK_SUBPORT_OWNER}})

        self.mock_trunk_get_object.called_once_with(mock.ANY, id=TRUNK_ID)
        TestTrunk.update.called_once_with(
            status=bc.trunk_consts.ACTIVE_STATUS)
        self.mock_trunk_get_object.assert_called_once_with(
            mock.ANY, id=TRUNK_ID)

    def test_is_trunk_parentport(self):
        return_value = self.handler.is_trunk_parentport(PORT_VM)

        self.assertTrue(return_value)

    def test_is_trunk_parentport_no_trunk(self):
        PORT_VM_NO_TRUNK = PORT_VM.copy()
        del PORT_VM_NO_TRUNK['trunk_details']
        return_value = self.handler.is_trunk_parentport(PORT_VM_NO_TRUNK)

        self.assertFalse(return_value)

    def test_is_trunk_subport(self):
        PORT_VM['device_owner'] = bc.trunk_consts.TRUNK_SUBPORT_OWNER
        return_value = self.handler.is_trunk_subport(PORT_VM)

        self.assertTrue(return_value)

    def test_is_trunk_subport_invalid_deviceowner(self):
        PORT_VM['device_owner'] = 'fake_owner'
        return_value = self.handler.is_trunk_subport(PORT_VM)

        self.assertFalse(return_value)

    def test_update_subports_baremetal(self):
        self._test_update_subports(PORT_BAREMETAL, DNS_NAME)

    def test_is_trunk_subport_baremetal(self):
        self.plugin.get_port.return_value = PORT_BAREMETAL
        return_value = self.handler.is_trunk_subport_baremetal(PORT_BAREMETAL)

        self.assertTrue(return_value)
        self.mock_subport_get_object.assert_called_once_with(
            mock.ANY, port_id=PORT_BAREMETAL['id'])
        self.mock_trunk_get_object.assert_called_once_with(
            mock.ANY, id=TestSubPort().trunk_id)

    def test_is_trunk_subport_baremetal_no_subport(self):
        self.mock_subport_get_object.return_value = None
        return_value = self.handler.is_trunk_subport_baremetal(PORT_BAREMETAL)

        self.assertFalse(return_value)
        self.mock_subport_get_object.assert_called_once_with(
            mock.ANY, port_id=PORT_BAREMETAL['id'])
        self.assertFalse(self.mock_trunk_get_object.call_count)

    def test_is_trunk_subport_baremetal_vm_port(self):
        self.plugin.get_port.return_value = PORT_VM
        return_value = self.handler.is_trunk_subport_baremetal(PORT_VM)

        self.assertFalse(return_value)
